import copy
import json
import os
from os.path import join

import yaml
from sentry_sdk import capture_exception


class ConfigAgent():
    """
    Class to configure the initial values of the meili_agent
    """

    def __init__(self, log_info, log_error, file_name_vehicles, file_name_topics ):
        # Logging
        self.log_info = log_info
        self.log_error = log_error


        # config agent by reading and storing config files about
        # number of vehicles, prefixes and topics
        self.file_name_vehicles = file_name_vehicles
        self.file_name_topics = file_name_topics
        self.cfg_file_path = os.path.join(os.path.expanduser("~"), ".meili", "cfg.yaml")

        self.data = None
        self.topics = None
        self.cfg = None

    def check_empty(self, variable, name):
        if not variable:
            self.log_error("[Config] %s is empty" % (name,))
            capture_exception(AssertionError)
            raise AssertionError("[Config] Empty", name)

    def open_yaml(self, filename):
        try:
            with open(filename, "r", encoding="utf-8") as config:
                return yaml.safe_load(config)
        except OSError:
            self.log_error("[Config] %s does not exist" % (filename,))
            capture_exception(OSError)
            raise

    def open_json(self, filename):
        try:
            with open(filename, "r", encoding="utf-8") as t:
                return json.load(t)
        except OSError:
            self.log_error("[Config] %s does not exist" % (filename,))
            capture_exception(OSError)
            raise


    def open_files(self):
        # Parse Mode, number of vehicles and tokens from config files passed as arguments
        self.topics = self.open_json(self.file_name_topics)
        self.data = self.open_yaml(self.file_name_vehicles)
        self.cfg = self.open_yaml(self.cfg_file_path)

        self.check_empty(self.topics, "Topics file")
        self.check_empty(self.data, "Data file")
        self.check_empty(self.cfg, "Config file")

    def config_var(self):
        try:
            setup_token = self.cfg["token"]
            server_instance = self.cfg["site"]
            token = self.data["fleet"]

            mqtt_id = self.data["mqtt_id"]

            fleet = True
        except KeyError as e:
            self.log_error(
                "[Config] Configuration of variable %s error" % (e,)
            )
            capture_exception(e)
            raise

        self.check_empty(setup_token, "Config file, token")
        self.check_empty(server_instance, "Config file, server_instance")
        self.check_empty(token, "Data file, fleet")
        self.check_empty(mqtt_id, "Data file, mqtt_id")

        mode = "Fleet"
        self.log_info("[Config] Fleet mode is : %s" % (mode,))

        return setup_token, server_instance, token, mqtt_id, fleet

    def set_vehicles_topics(self, vehicles):
        try:
            for vehicle in vehicles:
                t = copy.deepcopy(self.topics.get("topics", {}))
                t.update(vehicle.get("topics", {}))
                for topic in t:
                    if (vehicle["prefix"]) is not None:
                        t[topic]["topic"] = "{}{}".format(
                            vehicle["prefix"], t[topic]["topic"]
                        )
                    else:
                        t[topic]["topic"] = "{}".format(t[topic]["topic"])

                vehicle["topics"] = t
            return vehicles

        except TypeError:
            self.log_error("[Config] Error setting the vehicles topics")
            capture_exception(TypeError)
            raise

    def config_vehicles(self):

        self.check_empty(self.data["vehicles"], "Data file, vehicles")
        vehicles = self.set_vehicles_topics(self.data["vehicles"])

        vehicle_list = []
        vehicle_tokens = []
        vehicle_names = []

        try:
            # creating sublist of vehicles and tokens
            for vehicle in vehicles:
                vehicle_list.append(vehicle["uuid"])
                vehicle_tokens.append(vehicle["token"])
                vehicle_name = vehicle["token"].split(" ")[0]
                vehicle_names.append(vehicle_name)

        except KeyError as error:
            self.log_error(
                "[Config] Configuration of variable %s error" % (error,)
            )
            capture_exception(error)
            raise

        return vehicle_list, vehicles, vehicle_tokens, vehicle_names



from os import environ
from meili_ros_lib.sentry import agent_sentry
from numpy.distutils.fcompiler import str2bool


class Setup:
    def __init__(self):
        # Variables initialization
        self.aws_access_key_id = None
        self.aws_secret_access_key = None
        self.path_planning = None
        self.traffic_control = None
        self.battery_present = False
        self.lipo_battery = None
        self.outdoor = False
        self.topic_streaming_mode = None
        self.publish_frequency = None
        self.offline_flag = None

        # Logging
        self.log_info = None
        self.log_error = None

    def init_parameters_setup(self):
        try:

            if (
                    environ.get("AWS_ACCESS_KEY_ID") is not None
                    and environ.get("AWS_SECRET_ACCESS_KEY") is not None
            ):
                self.aws_access_key_id = environ.get("AWS_ACCESS_KEY_ID")
                self.aws_secret_access_key = environ.get("AWS_SECRET_ACCESS_KEY")
            else:
                self.aws_access_key_id = None
                self.aws_secret_access_key = None

            # get ROS distro
            ros_distro = environ["ROS_DISTRO"]
            ros_version = environ["ROS_VERSION"]

            self.log_info("[SetUp] Reading initialization parameters")
            self.log_info(f"[SetUp] Ros distro is : {ros_distro}")
            self.log_info(f"[SetUp] Ros version is :{ros_version}")

        except KeyError:
            self.log_error(f"[SetUp] Parameter {KeyError} does not exist. ")
            agent_sentry(KeyError)
            self.log_error("[SetUp] Roscore is not running, check your ros config")
            raise

        self.log_info(
            "[SetUp] COMPLETED Initial ROS Parameter configuration"
        )

        # get initial ros parameters
    def battery_parameter(self, battery):
        battery = str2bool(battery)
        if battery == "lipo":
            self.battery_present = True
            self.lipo_battery = True
            self.log_info("[SetUp] The Lipo battery is present")

        elif battery:
            self.battery_present = True
            self.log_info("[SetUp] The Battery is present")




import urllib

from meili_ros_lib.offline import OfflineAgent
from sentry_sdk import capture_exception, capture_message


class Connection:
    def __init__(
            self,
            agent_node,
            handlers,
            vehicle_list,
            disconnection_buffer=5

    ):
        self.offline_flag = agent_node.offline_flag
        self.log_info = agent_node.log_info
        self.log_error = agent_node.log_error
        self.handlers = handlers
        self.vehicle_list = vehicle_list

        self.client = None
        self.internet_down = 0
        self.offline_agent = OfflineAgent(vehicle_list, handlers, self.log_info, self.log_error)

        self.disconnection_buffer = disconnection_buffer

        self.log_info("[OfflineAgent] Initialized agent connection")

    def no_internet(self, ws_open):
        # The internet must be done disc_buffer (disc_buffer) times consecutive in a row
        # This avoids confusion with short internet disconnections
        try:
            if self.internet_down != self.disconnection_buffer:
                self.internet_down += 1
                self.log_info(
                    f"[OfflineAgent] Agent disconnected {self.internet_down}/{self.disconnection_buffer}")
            else:
                self.offline_agent.offline = True

                if ws_open:
                    # self.close_ws()
                    ws_open = False

                if self.offline_flag:
                    if self.offline_agent.offline_tasks is None:
                        self.offline_agent.offline_tasks = self.offline_agent.config_offlinetasks()
                    else:
                        self.offline_agent.activate_offline_agent()
        except Exception as e:
            self.log_info(f"ERROR >> {e}")
        return ws_open

    def check_internet_connection(self, ws_open, host='https://demo.meilirobots.com'):

        try:
            # CHECK FOR INTERNET CONNECTION
            urllib.request.urlopen(host, timeout=5)  # nosec
            self.offline_agent.offline = False
            self.internet_down = 0

            if ws_open is not True:
                self.open_ws()
                ws_open = True
            else:
                # IF WS IS ALREADY OPEN SEND OFFLINE LOG
                pass
                # TODO Send Offline Log in SDK
                # if self.ws.send_offlinelog():
                #    rospy.loginfo("[OfflineAgent] Offline Log Task file sent")
            return ws_open

        except Exception as e:
            ws_open = self.no_internet(ws_open)

            return ws_open

    def open_ws(self):
        self.log_info("[OfflineAgent] Internet Connected")
        # START WS IF IT IS NOT ALREADY OPEN
        try:
            self.client = self.handlers.sdk_setup()
            for index in self.vehicle_list:
                try:
                    self.client.add_vehicle(index)
                except Exception as e:
                    if e == "Vehicle already exists":
                        continue
            self.client.run_in_thread()

        except Exception as e:
            capture_exception(e)
            capture_message(f"[OfflineAgent] Error configuring ws: {e}")
            self.log_error(f"[OfflineAgent] Error configuring ws: {e}")



from meili_sdk.websockets.client import MeiliWebsocketClient
from sentry_sdk import capture_exception, capture_message
from meili_ros_lib.parse_data import parse_waypoints, parse_waypoints_v2, parse_docking_routine
from meili_ros_lib.sentry import agent_sentry

from std_msgs.msg import Bool

import time


class Handlers:

    def __init__(self, agent):

        self.agent = agent
        self.client = []  # docking routine

        for i in range(0, self.agent.number_of_vehicles):
            self.client.append(0)  # docking routine

        self.rate = None

    def sdk_setup(self):
        """sdk_setup"""
        try:
            self.agent.client = MeiliWebsocketClient(
                self.agent.token,
                override_host=self.agent.server_instance.replace("http", "ws"),
                fleet=self.agent.fleet,
                open_handler=self.open_handler,
                close_handler=self.close_handler,
                error_handler=self.error_handler,
                task_v2_handler=self.task_v2_handler,
                task_handler=self.task_handler,
                task_cancellation_handler=self.task_cancellation_handler,
                move_action_handler=self.move_action_handler,
                slow_down_handler=self.slow_down_handler,
                topic_list_initializer_handler=self.topic_handler,
                path_rerouting_handler=self.path_rerouting_handler,
                docking_routine_request_handler=self.docking_routine_request_handler,
                docking_routine_finalize_handler=self.docking_routine_finalize_handler,
                collision_clearance_handler=self.collision_clearance_handler,
                update_map_handler=self.update_map_handler,
            )
        except KeyError as error:
            self.agent.log_info(f"[ROSHandler] No handler named {error}")
            capture_exception(error)
        return self.agent.client

    def check_fleet_get_vehicle_position(self, vehicle):
        if self.agent.fleet:
            return self.agent.return_vehicle_position(vehicle)
        return 0

    def move_action_handler(self, move_msg, data: dict, vehicle: str):
        """When a task is received, goal is generated and sent to the specific robot"""
        try:
            vehicle_position = self.check_fleet_get_vehicle_position(vehicle)
            # For v2 testing--> will deprecate it
            if "metric_waypoints" in data["data"]["location"]:
                data = data["data"]["location"]
                pose = parse_waypoints(data)
                current_task_uuid = "uuid"
                self.waypoints_handler(
                    vehicle_position=vehicle_position,
                    pose=pose,
                )
            else:
                self.send_goal(data, vehicle_position)

            self.rate.sleep()

            # If the robot is moving the task will be automatically cancelled to go to the
            # charging station.
            self.agent.log_info(
                f"[ROSHandler] Sending vehicle {self.agent.vehicle_list[vehicle_position]}: "
                f"{self.agent.vehicle_names[vehicle_position]} to RESTING/CHARGING STATION"
            )

        except Exception as error:
            self.agent.log_error(
                "[ROSHandler] Error move_action_handler: %s" % (error,)
            )
            agent_sentry(error, "[ROSHandler] Error move_action_handler")

    def task_handler(self, _, data: dict, vehicle: str):
        """When a task is received, goal is generated and sent to the specific robot"""
        if self.agent.node.v2:
            return 0
        else:
            try:
                vehicle_position = self.check_fleet_get_vehicle_position(vehicle)
                current_task_uuid = data["uuid"]

                self.agent.number_of_tasks = self.agent.number_of_tasks + 1

                self.agent.log_info(
                    f"[ROSHandler] Number of received task is: {self.agent.number_of_tasks}"
                )
                self.agent.log_info(
                    f"[ROSHandler] Received task assigned to vehicle: {self.agent.vehicle_names[vehicle_position]}"
                )
                if not data["metric_waypoints"]:
                    # Case when only one task is sent
                    self.send_goal(data, vehicle_position)
                else:
                    # This is batch-task vehicle
                    if self.agent.node.path_planning:
                        pose = parse_waypoints(data)
                        # self.agent.total_waypoints = len(x_meters)
                        self.waypoints_handler(
                            vehicle_position=vehicle_position,
                            pose=pose,
                        )
                    else:
                        self.send_goal(data, vehicle_position)

                self.rate.sleep()

            except KeyError as error:
                self.agent.log_error(
                    f"[ROSHandler] Error task_handler: {error}"
                )
                agent_sentry(error, "[ROSHandler] Error task_handler")

    def task_v2_handler(self, task, vehicle_uuid: str):
        """When a task is received, goal is generated and sent to the specific robot"""
        if not self.agent.node.v2:
            return 0
        else:
            try:
                vehicle_position = self.check_fleet_get_vehicle_position(vehicle_uuid)
                self.agent.number_of_tasks = self.agent.number_of_tasks + 1

                self.agent.log_info(
                    f"[ROSHandler] Number of received v2 task is: {self.agent.number_of_tasks}"
                )
                self.agent.log_info(
                    f"[ROSHandler] Received v2 task assigned to vehicle: {self.agent.vehicle_names[vehicle_position]}"
                )
                self.action_handler(task, vehicle_uuid)
                self.rate.sleep()

            except KeyError as error:
                self.agent.log_error(
                    f"[ROSHandler] Error task v2_handler: {error}"
                )
                agent_sentry(error, "[ROSHandler] Error task_v2handler")
    
    def action_handler(self, task, vehicle_uuid):
        vehicle_position = self.check_fleet_get_vehicle_position(vehicle_uuid)
        action_type=task.action.action_type
        point=task.action.point
        form_values = task.action.values
        metric_waypoints=task.metric_waypoints
        rotation_angles= task.rotation_angles

        if (action_type=="move_to_point" or action_type=="go_to_charge" or action_type=="follow_path"):
            if metric_waypoints:
                pose = parse_waypoints_v2(metric_waypoints, rotation_angles)
                self.waypoints_handler(
                    vehicle_position=vehicle_position,
                    pose=pose,
                )
            else:
                self.send_goal_v2(point, vehicle_position)

        #Handling actions TO BE COMPLETED BY USER FOR THE SPECIFICS ACTIONS
        else:
            goal_id=task.uuid
            status_id=1
            self.agent.action_in_progress(goal_id, status_id, vehicle_uuid)
            time.sleep(0.05)
            status_id=3
            self.agent.action_finished(goal_id, status_id, vehicle_uuid)

    def task_cancellation_handler(self, cancel_msg, vehicle):
        """Handles task cancellation info to specific robot"""
        goal_id=cancel_msg.goal_id
        vehicle_position = self.check_fleet_get_vehicle_position(vehicle)
        self.agent.log_info(
            f"[ROSHandler] Cancelling task with of vehicle: {self.agent.vehicle_names[vehicle_position]} goal id: {goal_id}")

        # if cancellation is received cancel goal
        # Refresh variables (it has to be before the cancel_goal so the path stops being sent and there is no delay
        # for the traffic control)
        if self.agent.waypoints[vehicle_position]:
            # This line should be first to cancel the waypoints thread in ROS1
            self.agent.waypoints[vehicle_position] = False
            self.cancel_goal_waypoints(goal_id, vehicle_position)
        else:
            self.cancel_goal(goal_id, vehicle_position)

    def waypoints_handler(
            self, vehicle_position, pose
    ):
        """Handles offline waypoints and sends to specific robot"""
        # ROS2
        try:
            self.agent.waypoints[vehicle_position] = True
            self.send_waypoints(pose, vehicle_position)
        except KeyError as error:
            self.agent.log_error(
                f"[ROSHandler] Error waypoints_handler:{error}"
            )

            agent_sentry(error, "[ROSHandler] Error waypoints_handler")

    # def get_list_of_topic():
    #     """Getting the list of topics available with specified namespace"""
    #     # getting  topics available
    #     topic_list = rospy.get_published_topics(namespace="/robot1")
    #     return topic_list

    def topic_request_handler(self, _):
        """logging information"""
        self.agent.log_info("[ROSHandler] Received topic request")
        # get_list_of_topic()

    def close_handler(self, *_):
        """Close the websocket"""
        self.agent.log_info("[ROSHandler] WS close for Agent")
        self.agent.ws_open = False
        capture_message("[ROSHandler] WS close for Agent")

    def error_handler(self, _, err):
        """verifies websocket connection"""
        if not self.agent.connection.offline_agent.offline:
            self.agent.log_error(f"[ROSHandler] WS error {err}")
            capture_message("[ROSHandler] WS error")

    def open_handler(self, *_, **__):
        """Opens the websocket connection"""
        self.agent.log_info("[ROSHandler] WS open")
        self.agent.ws_open = True

    def path_rerouting_handler(self, path_rerouting_msg, data: dict, vehicle: str):
        path=path_rerouting_msg["path"]
        vehicle_position = self.check_fleet_get_vehicle_position(vehicle)
        
        if path is not None:
            self.agent.pause_paths[vehicle_position] = True
            self.agent.log_info(f"[ROSHandler] Path Rerouting vehicle: {self.agent.vehicle_names[vehicle_position]}")
            self.agent.goal_id_rerouted[vehicle_position] = self.agent.goal_id[vehicle_position]
            self.task_cancellation_handler(goal_id=self.agent.goal_id_rerouted[vehicle_position], vehicle=vehicle)

            data_new_path = {"rotation_angles": data["rotation_angles"], "metric_waypoints": data["path"]}
            current_task_uuid = "Rerouted_task"

            pose = parse_waypoints(data_new_path)
            self.agent.waypoints[vehicle_position] = True # To Test
            self.waypoints_handler(
                vehicle_position=vehicle_position,
                pose=pose,
            )
        else:
            self.agent.log_error("[ROSHandler] Received empty path from path rerouting message.No action is done.")


    def docking_routine_finalize_handler(self, dock_msg: dict, vehicle: str):
        self.agent.log_info("[ROSHandler] Docking Routine FINALIZING request")
        vehicle_position = self.check_fleet_get_vehicle_position(vehicle)

        try:
            vehicle = self.agent.vehicles[vehicle_position]
            vehicle_prefix = str(vehicle["prefix"])
            if vehicle_prefix == "None":
                vehicle_prefix = ""

            pub = self.end_recording_publisher(vehicle_prefix)

            for i in range(10):
                variable = Bool()
                variable.data = False
                pub.publish(variable)
                time.sleep(1)
            self.agent.log_info("[ROSHandler] Published end of recording")

            self.client[vehicle_position].wait_for_result()
            result = self.client[vehicle_position].get_result()

            if result is not None:
                # rospy.loginfo(f"[ROSHandler] Result {result}")
                path, rotation_angles = parse_docking_routine(result.plan)
                self.agent.ping_docking_routine(vehicle["uuid"], path, rotation_angles)
            else:
                self.agent.log_error(f"[ROSHandler] Docking output is None. Try again.")
        except Exception as e:
            self.agent.log_error(f"[ROSHandler] Error on Docking Routine Finalize: {e}")

    def slow_down_handler(self, slow_down_msg, data: dict, vehicle: str):
        # Method should be written in the particular ROS Handler
        pass

    def docking_routine_request_handler(self, dock_msg: dict, vehicle: str):
        # Method should be written in the particular ROS Handler
        pass

    def collision_clearance_handler(self, collision_clearance_msg: dict, vehicle: str):
        # Method should be written in the particular ROS Handler
        pass

    def update_map_handler(self, data: dict, vehicle: str):
        # Method should be written in the particular ROS Handler
        pass



import csv
import datetime
from os.path import expanduser, isfile, join

import numpy
import yaml
from sentry_sdk import capture_exception, capture_message


def config_time(task):
    time = [task["time"]]
    for i in range(len(time)):
        if len(time[i]) != 5:
            raise AssertionError("[OfflineAgent] Time data incorrect length")

    try:
        t = time[0].split(":")
        total_minutes = int(t[0]) * 60 + int(t[1]) * 1

    except BaseException as e:
        capture_exception(e)
        capture_message("[OfflineAgent] Time data incorrect")
        raise BaseException from e

    cron_split = []
    cron = 0

    if task["cron"] != "":
        cron_split.append(task["cron"].split())
        cron_raw = cron_split
        cron_time1 = cron_raw[0][0]
        if len(cron_time1) < 5:
            if len(cron_time1) == 1:
                cron = 0
            elif len(cron_time1) == 4:
                cron_str = cron_time1[2] + cron_time1[3]
                cron = int(cron_str)
            else:
                cron_str = cron_time1[2]
                cron = int(cron_str)

        elif len(cron_time1) > 5:
            if len(cron_time1) == 7:
                cron_str = cron_time1[5] + cron_time1[6]
                cron = int(cron_str)
            elif len(cron_time1) == 6:
                cron_str = cron_time1[5]
                cron = int(cron_str)
        else:
            cron = 0

        new_time = total_minutes

        while cron != 0 and new_time < 1440:
            new_time = new_time + cron
            if new_time > 1440:
                new_time = 1440
            new_time_datetime = "{:02d}:{:02d}".format(*divmod(new_time, 60))
            time.append(new_time_datetime)
    else:
        cron_split.append("")

    return time


class OfflineAgent:
    def __init__(
            self, vehicle_list, handlers, log_info, log_error

    ):

        self.log_info = log_info
        self.log_error = log_error
        self.vehicle_list = vehicle_list
        self.number_of_vehicles = len(vehicle_list)
        self.handlers = handlers

        self.current_task_uuid = None
        self.task_offline_started = False
        self.last_executed_time = None

        self.offline = None
        self.offline_tasks = None

    def activate_offline_agent(self):
        #self.log_info("[OfflineAgent] Offline Agent Activated")

        self.task_offline_started= self.offline_task_executer(
            self.offline_tasks,
            self.vehicle_list,
            self.task_offline_started,
        )
        self.offline = True

    def config_weekdays(self, task):
        week_days = numpy.zeros(7, dtype=int)
        weeks_days = []

        for day in task["week_days"]:
            if not isinstance(day, int):
                raise TypeError("[OfflineAgent] Day is not integer")

            if int(day) < 0 or int(day) > 7:
                raise TypeError("[OfflineAgent] Day must be between 0 to 6")

            week_days[int(day)] = 1
            weeks_days.append(int(day))

        # If no days are specified it is done for all the days
        if len(weeks_days) == 0:
            self.log_info(
                "[OfflineAgent] Offline task where no weekdays where specified"
            )
            for day in range(7):
                week_days[int(day)] = 1
                weeks_days.append(int(day))

        return weeks_days

    def config_goal(self, task):
        goals = []
        for goal in task["task_preset"]["subtasks"]:
            point = goal[list(goal.keys())[1]]
            point_type = point["point_type"]

            if not isinstance(point_type, str):
                raise TypeError("[OfflineAgent] Point Type is not string")
            # self.check_empty(point_type, "Offline Task, point_type")
            x = point["x_coordinate"]
            y = point["y_coordinate"]
            self.check_empty(x, "Offline Task, x")
            self.check_empty(y, "Offline Task, y")
            rotation = point["rotation"]
            self.check_empty(rotation, "Offline Task, rotation")
            if not isinstance(rotation, int) or (rotation > 361 or rotation < 0):
                raise TypeError("[OfflineAgent] Rotation is an invalid data")

            goals.append([x, y, rotation])

        return goals

    def config_loc_uuid(self, task):
        location_uuid = []

        for goal in task["task_preset"]["subtasks"]:

            point = goal[list(goal.keys())[1]]

            if not isinstance(point["uuid"], str):
                raise TypeError("[OfflineAgent] Uuid is not string")

            self.check_empty(point["uuid"], "Offline Task, point_uuid")

            location_uuid.append(point["uuid"])

        return location_uuid

    def open_yaml(self, filename):
        try:
            with open(filename, "r") as config:
                return yaml.safe_load(config)
        except OSError:
            capture_exception(OSError)

            self.log_info(
                "[OfflineAgent] %s does not exist" % (filename,)
            )

    def config_offlinetasks(self):
        offline_tasks = []
        self.log_info("[OfflineAgent] Configuring Offline Tasks")

        if self.offline:
            try:
                tasks_file_path = join(expanduser("~"), ".meili/tasks.yaml")
                data = self.open_yaml(tasks_file_path)
                self.check_empty(data, "Offline Tasks file")

                self.last_executed_time = [0 for _ in range(self.number_of_vehicles)]

                for task in data:
                    uuid = task["uuid"]
                    vehicle = task["vehicle"]["uuid"]

                    if not isinstance(uuid, str):
                        raise TypeError("[OfflineAgent] Offline Task file, task uuid")
                    if not isinstance(vehicle, str):
                        raise TypeError(
                            "[OfflineAgent] Offline Task file, vehicle uuid"
                        )
                    self.check_empty(uuid, "Offline Task file, uuid")
                    self.check_empty(vehicle, "Offline Task file, vehicle uuid")

                    goals = []
                    weeks_days = []
                    time = []
                    location_uuid = []
                    try:
                        weeks_days = self.config_weekdays(task)
                        goals = self.config_goal(task)
                        location_uuid = self.config_loc_uuid(task)
                        time = config_time(task)
                    except Exception as e:
                        self.log_error(
                            "[OfflineAgent] Configuration Error: %s" % e,
                        )

                    if len(vehicle) != 0 and len(goals) != 0:
                        offline_tasks.append(
                            [uuid, weeks_days, time, vehicle, goals, location_uuid]
                        )

                self.log_info("[OfflineAgent] Offline Tasks Configured")
            except AssertionError:
                capture_exception(AssertionError)
                capture_message("[OfflineAgent] Error Configuring Offline Tasks")
                self.log_error(
                    "[OfflineAgent] Error Configuring Offline Tasks"
                )
                exit()

        return offline_tasks

    def get_vehicle_position(self, vehicle_list, vehicle_uuid):
        try:
            position = vehicle_list.index(vehicle_uuid)
        except ValueError:
            capture_exception(ValueError)
            self.log_info(
                "[OfflineAgent] Vehicle %s is not in the list of vehicle, task will be sent to the first one in the "
                "list "
                % (vehicle_uuid,)
            )
            position = 0

        return position

    def offline_task_handler(self, pose, location, vehicle_position):
        if len(pose) == 1:
            x = pose[0][1]
            y = pose[0][2]
            xm = x
            ym = y
            rotation = pose[0][3]
            vehicle = self.offline_tasks[0][3]

            task = {
                "location": location,
                "number": location,
                "uuid": location,
            }
            data = {
                "uuid": location,
                "number": location,
                "location": {
                    "uuid": location,
                    "x": x,
                    "y": y,
                    "metric": {"x": xm, "y": ym},
                    "rotation": rotation,
                },
                "metric_waypoints": None,
            }

            self.handlers.task_handler(
                task=task,
                data=data,
                vehicle=vehicle,
            )
        else:
            self.handlers.waypoints_handler(vehicle_position=vehicle_position, pose=pose, current_task_uuid=location)

    def offline_task_executer(
            self,
            offline_tasks,
            vehicle_list,
            task_offline_started,
    ):
        today = datetime.date.today().weekday()
        now = datetime.datetime.now()
        current_time = now.strftime("%H:%M")

        self.offline_tasks = offline_tasks

        for task in offline_tasks:
            task_execution_time = task[2]
            task_execution_day = task[1]

            pose = task[4]
            vehicle_uuid = task[3]
            location = task[5]

            position = self.get_vehicle_position(vehicle_list, vehicle_uuid)

            for day in task_execution_day:
                if today == day:
                    for execution_time in task_execution_time:
                        if (
                                current_time == execution_time
                                and task_offline_started is False
                                and current_time != self.last_executed_time[position]
                        ):
                            task_offline_started = True
                            self.last_executed_time[position] = execution_time
                            self.current_task_uuid = task[0]
                            self.offline_task_handler(pose, location, position)

        return task_offline_started

    def logging_offlinetasks(
            self, offline_task, number_of_tasks, vehicle_list, start_time, success
    ):

        task_number = number_of_tasks
        uuid = offline_task[0]

        # TO DO: add time+date
        vehicle = offline_task[3]
        location_uuid = offline_task[5]
        time = self.last_executed_time[vehicle_list.index(vehicle)]
        now = datetime.datetime.now()
        end_time = now.strftime("%d/%m/%Y %H:%M:%S")

        tasks_log_file_path = join(expanduser("~"), ".meili/log_tasks.csv")
        file_exists = isfile(tasks_log_file_path)

        with open(tasks_log_file_path, "a+", newline="") as file:
            fieldnames = [
                "no",
                "time",
                "uuid",
                "vehicle",
                "success",
                "location",
                "start_time",
                "end_time",
            ]
            writer = csv.DictWriter(file, fieldnames=fieldnames)
            if not file_exists:
                writer.writeheader()
            data = {
                "no": task_number,
                "time": time,
                "uuid": uuid,
                "vehicle": vehicle,
                "success": success,
                "location": location_uuid,
                "start_time": start_time,
                "end_time": end_time,
            }

            writer.writerow(data)
        self.log_info(
            "[ROSAgent] Task %s Logged in %s"
            % (
                task_number,
                tasks_log_file_path,
            )
        )

    def check_empty(self, variable, name):
        if not variable and variable != 0:
            self.log_error("[OfflineAgent] %s is empty" % (name,))
            raise AssertionError("[OfflineAgent] Empty", name)



import sys
import time
import datetime
import threading
from math import sqrt, dist

from meili_sdk.websockets.models.message import Message
from meili_sdk.websockets import constants

from geometry_msgs.msg import PoseWithCovarianceStamped, Pose, Twist
from sensor_msgs.msg import NavSatFix
from nav_msgs.msg import Odometry

from meili_ros_lib.config_agent import ConfigAgent
from meili_ros_lib.maths import quaternion_to_euler
from meili_ros_lib.parse_data import parse_battery_data
from meili_ros_lib.sentry import initialize_sentry, agent_sentry

from sentry_sdk import capture_exception, capture_message


initialize_sentry()


def update_capacity_lipo(voltage):
    # Calculate current capacity of lipo batteries taking into account voltage/capacity estimation using regression
    capacity = 47.6 * voltage - 500
    return capacity


class Agent:
    def __init__(self, node, file_name_vehicles, file_name_topics):
        # Logging functions

        # Threads
        self.thread1 = None
        self.thread2 = None

        self.log_info = node.log_info
        self.log_error = node.log_error
        self.file_name_vehicles = file_name_vehicles
        self.file_name_topics = file_name_topics

        # Initial instances
        self.node = node
        self.token = None
        self.fleet = None
        self.mqtt_id = None

        self.vehicle_list = []
        self.vehicle_tokens = []
        self.vehicles = {}
        self.number_of_vehicles = None

        # Other classes
        self.connection = None
        self.client = None

        # Messages
        self.msg_pose = []
        self.msg_battery = []
        self.msg_gps = []
        self.msg_speed = []

        self.msg_plan = []
        self.arr_plan = []
        self.buffer = []
        # Internet
        self.ws_open = False

        # Task
        self.task_started = []
        self.number_of_tasks = 0
        self.start_time = None

        # Waypoints
        self.waypoints = []
        self.waypoints_goal_id = []

        # Traffic control
        self.goal_id = []
        self.goal_id_rerouted = []
        self.max_vel_x = []
        self.max_vel_theta = []
        self.path_length = 5  # in Meters
        self.pause_paths = {}

        # Miscellaneous
        self.total_topics = 0
        self.vehicles = self.__agent_setup()

    def __agent_setup(self):
        config = ConfigAgent(self.log_info, self.log_error, self.file_name_vehicles, self.file_name_topics)
        self.log_info("[ROSAgent] The configuration is started...")

        try:
            config.open_files()

            (
                setup_token,
                self.server_instance,
                self.token,
                self.mqtt_id,
                self.fleet,
            ) = config.config_var()

            self.log_info(f"[ROSAgent] Server instance:{self.server_instance}")
            (
                self.vehicle_list,
                vehicles,
                self.vehicle_tokens,
                self.vehicle_names
            ) = config.config_vehicles()

            self.number_of_vehicles = len(self.vehicle_list)
            self.initialisation()

        except Exception as error:
            self.log_error(
                "[ROSAgent] Agent Setup Error: %s %s" % (type(error).__name__, error)
            )
            agent_sentry(error, "[ROSAgent] Agent Setup Error")
            sys.exit()

        for vehicle in vehicles:
            uuid = vehicle["uuid"]
            self.log_info(f"[ROSAgent] Vehicle {uuid} is on the fleet")
        self.log_info("[ROSAgent] Agent has been configured")

        return vehicles

    def initialisation(self):
        for i in range(0, self.number_of_vehicles):
            self.msg_pose.append(i)
            self.msg_battery.append(i)
            self.msg_gps.append(i)
            self.msg_speed.append(i)

            self.msg_plan.append(i)
            self.arr_plan.append([])
            self.buffer.append(i)

            self.task_started.append(0)
            self.waypoints.append(False)
            self.waypoints_goal_id.append(None)
            self.goal_id_rerouted.append(None)
            self.pause_paths[i] = False

            self.max_vel_x.append(0)
            self.max_vel_theta.append(0)
            self.goal_id.append(0)

    ################################################
    def check_topic_streaming_mode(self):
        """Check Topic Streaming"""
        if self.node.topic_streaming_mode == "FILTERING":
            self.node.log_info("[ROSAgent] Mode for topic streaming is FILTERING")
        else:
            self.node.log_info("[ROSAgent] Mode for topic streaming is REALTIME")

    def check_internet(self):
        """Check if there is internet if not connect the offline agent"""
        try:
            self.ws_open = self.connection.check_internet_connection(self.ws_open)
        except Exception as error:
            self.node.log_info(f"[ROSAgent] Error Checking internet:{error}")

    ################################################

    # PING FUNCTIONS
    def send_message_client(self, event, value, vehicle):
        """Sends message via Websocket"""
        try:
            message = Message(event=event, value=value, vehicle=vehicle)
            message.validate()
            self.client.send(message)
        except Exception as e:
            self.log_info(f"SEND MESSAGE ERROR{e} {event} {vehicle}")

    def ping_filtering_mode(self):

        self.ping_speed()

        if self.node.outdoor:
            self.ping_gps()
        else:
            self.ping_location()

        if self.node.battery_present:
            self.ping_battery()

        if self.node.traffic_control and not self.node.v2:
            self.ping_trajectory()

    def ping_location(self):
        """Ping Location """
        try:
            for index, msg in enumerate(self.msg_pose):
                if isinstance(msg, PoseWithCovarianceStamped):
                    x = round(msg.pose.pose.position.x, 3)
                    y = round(msg.pose.pose.position.y, 3)
                    orientation = msg.pose.pose.orientation
                elif isinstance(msg, Pose):
                    x = round(msg.position.x, 3)
                    y = round(msg.position.y, 3)
                    orientation = msg.orientation
                else:
                    self.log_error(
                        f"[ROSAgent] Robot {index}-{self.vehicle_names[index]} pose is not available. Message >> {msg}"
                    )
                    capture_message("[ROSAgent] Robot pose is not available")
                    continue

                yaw = round(quaternion_to_euler(
                    orientation.x, orientation.y, orientation.z, orientation.w
                ),
                    3)
                self.send_message_client(
                    constants.EVENT_LOCATION,
                    {"timestamp": time.time(), "xm": x, "ym": y, "rotation": yaw},
                    self.vehicle_list[index],
                )

        except Exception as error:
            self.log_info(f"[ROSAgent] Ping location error: {error}")
            capture_exception(error)

    def ping_battery(self):
        """Sending Battery message via websocket"""
        try:
            for index, msg in enumerate(self.msg_battery):
                if self.node.lipo_battery:
                    msg.capacity = update_capacity_lipo(msg.voltage)
                battery_data = parse_battery_data(msg)
                battery_data["timestamp"] = time.time()
                self.send_message_client(
                    constants.EVENT_BATTERY,
                    {**battery_data},
                    self.vehicle_list[index],
                )
        except Exception as error:
            capture_exception(error)
            self.log_info(f"[ROSAgent] Ping Battery Error: {error}")

    def ping_trajectory(self):
        """Send the trajectory as an array of poses"""

        try:
            for index, msg in enumerate(self.arr_plan):
                if self.task_started[index] and not self.pause_paths[index]:
                    now = datetime.datetime.now()
                    current_time = now.strftime("%H:%M:%S")
                    # self.log_info(f"[ROSAgent] Sending path of vehicle {self.vehicle_names[index]}: {current_time}")
                    start = msg[0]
                    end = msg[-1]
                    dist = dist(start, end)
                    # self.log_info(f"{start}-- {end} ={dist} vehicle {self.vehicle_list[index]}")
                    self.send_message_client(
                        constants.EVENT_PATH_DATA,
                        {"timestamp": time.time(), "path": msg},
                        self.vehicle_list[index],
                    )
        except Exception as error:
            self.log_info(
                f"[ROSAgent] Ping Trajectory Error: {error} "
            )
            self.log_info(f"{index} {self.arr_plan}")
            sys.exit()
            capture_exception(error)

    def ping_gps(self):
        try:
            for index, msg in enumerate(self.msg_gps):
                if isinstance(msg, NavSatFix):
                    lat = msg.latitude
                    lon = msg.longitude
                    rotation = 0
                    self.send_message_client(
                        constants.EVENT_LOCATION,
                        {"timestamp": time.time(), "lat": lat, "lon": lon, "rotation": rotation},
                        self.vehicle_list[index],
                    )
        except Exception as e:
            self.log_info(
                "[ROSAgent] Ping GPS Error: %s ", e)
            capture_exception(e)

    def ping_docking_routine(self, vehicle, path, rotation_angles):
        # send the trajectory as an array of poses
        try:
            self.send_message_client(
                constants.EVENT_DOCKING_ROUTINE,
                {"timestamp": time.time(), "path": path, "rotation_angles": rotation_angles},
                vehicle,
            )
            self.log_info("[ROSAgent] Docking Routine Sent")
        except Exception as e:
            self.log_info(f"[ROSAgent] Ping Docking Routine Error: {e} ")
            capture_exception(e)

    def ping_speed(self):
        """Ping Speed"""
        try:
            for index, msg in enumerate(self.msg_speed):
                if isinstance(msg, Odometry):
                    speed = msg.twist.twist.linear.x
                    speed = round(speed, 3)
                    self.send_message_client(
                        constants.EVENT_SPEED,
                        {"timestamp": time.time(), "speed": speed},
                        self.vehicle_list[index]
                    )

                else:
                    #self.log_info(f"[ROSAgent] Robot {self.vehicle_names[index]} speed is not available")
                    #capture_message("[ROSAgent] Robot speed is not available")
                    break

        except Exception as e:
            self.log_info(f"[ROSAgent] Ping speed Error: {e} ")
            capture_exception(e)

    ################################################
    def return_vehicle_position(self, vehicle_token):
        """Returns: Vehicle Position"""
        return self.vehicle_list.index(vehicle_token)

    ################################################
    # CALLBACKS
    def callback_pose(self, msg, vehicle_uuid):
        """Parse current robot pose and sent to FMS task_msg_server"""
        vehicle_position = self.return_vehicle_position(vehicle_uuid)
        self.msg_pose[vehicle_position] = msg

        if self.node.topic_streaming_mode == "REALTIME " and self.ws_open:
            self.ping_location()

    def callback_battery(self, msg, vehicle_uuid):
        """When battery topic is available, is parsed and sent to FMS agent"""
        if self.node.battery_present:
            vehicle_position = self.return_vehicle_position(vehicle_uuid)
            self.msg_battery[vehicle_position] = msg

            if self.node.topic_streaming_mode == "REALTIME" and self.ws_open:
                self.ping_battery()

    def callback_rosoutagg(self, msg, type_msg):
        rosout_message = msg.msg

        count = 0
        # Find the vehicle uuid that corresponds to the message.
        if rosout_message.find('vehicle') != -1:
            for vehicle_uuid in self.vehicle_list:
                if (rosout_message.find(vehicle_uuid)) != -1:
                    vehicle = vehicle_uuid
        elif msg.name.find('robot') != -1:
            name = '/robot' + msg.name[msg.name.find('robot') + 5]
            for vehicle in self.vehicles:
                if (vehicle['prefix']) == name:
                    vehicle = self.vehicle_list[count]
                count += 1
        else:
            vehicle = self.vehicle_list[0]

        if self.ws_open and self.node.logging==type_msg:
            type_msg="warning"
            try:
                self.send_message_client(
                    constants.EVENT_NOTIFICATION,
                            {"vehicle": vehicle,
                             "message": rosout_message, "level": type_msg,
                             "meta": {"file": msg.file, "function": msg.function, "line": msg.line}},
                            vehicle,
                        )
            except Exception as e:
                self.log_error(f"[ROSAgent] Sending notification error {e}")

    def task_in_progress(self, goal_id, status_id, vehicle_uuid):
        if self.ws_open:
            vehicle_position = self.return_vehicle_position(vehicle_uuid)
            if self.waypoints_goal_id[vehicle_position] is not None:
                goal_id = self.waypoints_goal_id[vehicle_position]

                if self.goal_id_rerouted[vehicle_position] is not None:
                    goal_id = self.goal_id_rerouted[vehicle_position]

            self.goal_id[vehicle_position] = goal_id
            self.send_message_client(
                constants.EVENT_GOAL_STATUS,
                {"goal_id": goal_id, "status_id": status_id},
                vehicle_uuid,
            )
            #self.log_info(
            #    f"[ROSAgent] Sending {goal_id} in progress<<{status_id}>> of vehicle {vehicle_position} ")

        else:
            now = datetime.datetime.now()
            dt_string = now.strftime("%d/%m/%Y %H:%M:%S")
            self.start_time = dt_string
        return 1

    def action_in_progress(self, goal_id, status_id, vehicle_uuid):
        if self.ws_open:
            vehicle_position = self.return_vehicle_position(vehicle_uuid)
            self.goal_id[vehicle_position] = goal_id
            self.send_message_client(
                constants.EVENT_GOAL_STATUS,
                {"goal_id": goal_id, "status_id": status_id},
                vehicle_uuid,
            )
            self.log_info(
                f"[ROSAgent] Sending action {goal_id} in progress<<{status_id}>> of vehicle {vehicle_position} ")
        return 1

    def callback_status(self, msg, goal_id, status_id, vehicle_uuid):
        # Status of navigation task is parsed and send to FMS task_msg_server
        vehicle_position = self.return_vehicle_position(vehicle_uuid)

        if (
                self.waypoints_goal_id[vehicle_position] is None
                and self.waypoints[vehicle_position]
        ):
            self.waypoints_goal_id[vehicle_position] = goal_id
        # IN PROGRESS
        if status_id == 1 and self.task_started[vehicle_position] == 0:
            try:
                self.task_started[vehicle_position] = self.task_in_progress(
                    goal_id, status_id, vehicle_uuid
                )
            except Exception as e:
                self.log_error(
                    "[ROSAgent] Task in progress Callback Status error : %s", e
                )
                agent_sentry(e, "[ROSAgent] Task in progress Callback Status error")

        # PREEMPTED NEEDED
        if status_id == 2 and self.task_started[vehicle_position] == 1:
            self.task_started[vehicle_position] = self.task_preempted(goal_id, status_id, vehicle_uuid)

        # IF SUCCEEDED; ABORTED or REJECTED
        if (status_id in (3, 4, 5)) and self.task_started[vehicle_position] == 1:

            try:
                self.task_started[vehicle_position] = self.task_finished(
                    goal_id, status_id, vehicle_uuid
                )

            except Exception as e:
                self.log_error(
                    f"[ROSAgent] Task finished Callback Status error: {e}"
                )
                agent_sentry(e, "[ROSAgent] Task finished Callback Status error")

    def task_preempted(self, goal_id, status_id, vehicle_uuid):
        vehicle_position = self.return_vehicle_position(vehicle_uuid)
        if self.waypoints_goal_id[vehicle_position] is not None:
            goal_id = self.waypoints_goal_id[vehicle_position]
            self.waypoints_goal_id[vehicle_position] = None

        self.log_info(
            f"[ROSAgent] Task {goal_id} preempted <<{status_id}>> of vehicle {vehicle_position} ")

        return 0

    def task_finished(self, goal_id, status_id, vehicle_uuid):
        vehicle_position = self.return_vehicle_position(vehicle_uuid)
        # TASK ONLINE
        if self.ws_open:
           # self.log_info(
           #     f"[ROSAgent] Task {goal_id} finished<<{status_id}>> of vehicle {vehicle_position} ")
            if not self.waypoints[vehicle_position]:

                if self.waypoints_goal_id[vehicle_position] is not None:
                    goal_id = self.waypoints_goal_id[vehicle_position]
                    self.waypoints_goal_id[vehicle_position] = None
                    if self.goal_id_rerouted[vehicle_position] is not None:
                        goal_id = self.goal_id_rerouted[vehicle_position]
                        self.goal_id_rerouted[vehicle_position] = None

                self.goal_id[vehicle_position] = goal_id
                self.send_message_client(
                    constants.EVENT_GOAL_STATUS,
                    {"goal_id": goal_id, "status_id": status_id},
                    vehicle_uuid,
                )

              #  self.log_info(
              #      f"[ROSAgent] Sending {goal_id} finished<<{status_id}>> of vehicle {vehicle_position} ")

        # OFFLINE TASK
        if self.connection.offline_agent.offline:
            success = True

            for task in self.connection.offline_agent.offline_tasks:
                if self.connection.offline_agent.current_task_uuid == task[0]:

                    if not self.waypoints[vehicle_position]:
                        self.connection.offline_agent.logging_offlinetasks(
                            task,
                            self.number_of_tasks,
                            self.vehicle_list,
                            self.start_time,
                            success=success,
                        )
                        self.start_time = None
                        self.connection.offline_agent.task_offline_started = False
        return 0

    def action_finished(self, goal_id, status_id, vehicle_uuid):
        # TASK ONLINE
        if self.ws_open:
            vehicle_position = self.return_vehicle_position(vehicle_uuid)
            self.log_info(
                f"[ROSAgent] Task {goal_id} finished<<{status_id}>> of vehicle {vehicle_position} ")
            self.goal_id[vehicle_position] = goal_id
            self.send_message_client(
                constants.EVENT_GOAL_STATUS,
                {"goal_id": goal_id, "status_id": status_id},
                vehicle_uuid,
            )

            self.log_info(
                f"[ROSAgent] Sending action {goal_id} finished<<{status_id}>> of vehicle {vehicle_position} ")

    def check_distance_of_array(self, msg):
        path = []
        first = None
        for pose in msg.poses:
            if first is None:
                first = [pose.pose.position.x, pose.pose.position.y]
            point = [pose.pose.position.x, pose.pose.position.y]
            distance = dist(first, point)
            path.append(point)
            if distance > self.path_length:
                yield path
        yield path

    def callback_plan(self, msg, vehicle_uuid):
        """Callback Plan"""
        # Parse trajectory and send to the fms

        arr = next(self.check_distance_of_array(msg))
        vehicle_position = self.return_vehicle_position(vehicle_uuid)
        self.arr_plan[vehicle_position] = arr

        if (
                self.node.topic_streaming_mode == "REALTIME"
                and self.ws_open
        ):
            self.ping_trajectory()

    def callback_gps(self, msg, vehicle_uuid):
        """Callback GPS"""
        vehicle_position = self.return_vehicle_position(vehicle_uuid)
        self.msg_gps[vehicle_position] = msg
        # Parse current robot pose and sent to FMS task_msg_server
        if self.node.topic_streaming_mode == "REALTIME" and self.ws_open:
            self.ping_gps()

    def callback_speed(self, msg, vehicle_uuid):
        """ Callback Speed"""
        vehicle_position = self.return_vehicle_position(vehicle_uuid)
        self.msg_speed[vehicle_position] = msg
        # Parse current robot pose and sent to FMS task_msg_server
        if self.node.topic_streaming_mode == "REALTIME" and self.ws_open:
            self.ping_speed()

    ################################################
    def run(self, conn_rate, pub_rate):
        """Running the agent"""
        # When Mode is filtering, all data is sent to the FMS task_msg_server at specific frequency
        self.check_topic_streaming_mode()

        while True:
            try:
                # Check internet connection
                if not (self.thread1 and self.thread1.is_alive()):
                    self.thread1 = threading.Thread(target=self.connecting, args=(conn_rate,))
                    self.thread1.start()

                if not (self.thread2 and self.thread2.is_alive()):
                    self.thread2 = threading.Thread(target=self.pinging, args=(pub_rate,))
                    self.thread2.start()


            except Exception as error:
                capture_exception(error)
                self.log_error(
                    f"[ROSAgent] Run Error: {type(error).__name__}, {error}"
                )

    def connecting(self, conn_rate):
        if not self.connection.offline_agent.task_offline_started:
            self.check_internet()
            conn_rate.sleep()

    def pinging(self, pub_rate):
        if self.node.topic_streaming_mode == "FILTERING" and self.ws_open:
            self.ping_filtering_mode()
            pub_rate.sleep()



import os

import sentry_sdk
from sentry_sdk import capture_exception, capture_message

__all__ = ("initialize_sentry",)


def agent_sentry(exception=None, message=None):
    if message is not None:
        capture_message(message)
    if exception is not None:
        capture_exception(exception)


def initialize_sentry(key=None):
    """
    Initialize Sentry SDK

    Will only initialize sentry if key is provided or found in the configuration

    Set environemntal variable MEILI_SENTRY_DSN or ROS parameter /meili_sentry_dsn
    """
    key = key or os.environ.get("MEILI_SENTRY_DSN", None)
    if key:
        sentry_sdk.init(
            key,
            # Set traces_sample_rate to 1.0 to capture 100%
            # of transactions for performance monitoring.
            # We recommend adjusting this value in production.
            traces_sample_rate=1.0,
        )
